/*
			Copyright (c) 1993 by
			Advanced Visual Systems Inc.
			All Rights Reserved

	This software comprises unpublished confidential information of
	Advanced Visual Systems Inc. and may not be used, copied or made
	available to anyone, except in accordance with the license
	under which it is furnished.

	This file is under Perforce control
	$Id: //depot/express/fcs992/modules/mb_plt3d.c#1 $
*/
/****************************************************************************
                  INTERNATIONAL AVS CENTER
	(This disclaimer must remain at the top of all files)

WARRANTY DISCLAIMER

This module and the files associated with it are distributed free of charge.
It is placed in the public domain and permission is granted for anyone to use,
duplicate, modify, and redistribute it unless otherwise noted.  Some modules
may be copyrighted.  You agree to abide by the conditions also included in
the AVS Licensing Agreement, version 1.0, located in the main module
directory located at the International AVS Center ftp site and to include
the AVS Licensing Agreement when you distribute any files downloaded from 
that site.

The International AVS Center, MCNC, the AVS Consortium and the individual
submitting the module and files associated with said module provide absolutely
NO WARRANTY OF ANY KIND with respect to this software.  The entire risk as to
the quality and performance of this software is with the user.  IN NO EVENT
WILL The International AVS Center, MCNC, the AVS Consortium and the individual
submitting the module and files associated with said module BE LIABLE TO
ANYONE FOR ANY DAMAGES ARISING FROM THE USE OF THIS SOFTWARE, INCLUDING,
WITHOUT LIMITATION, DAMAGES RESULTING FROM LOST DATA OR LOST PROFITS, OR ANY
SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES.

This AVS module and associated files are public domain software unless
otherwise noted.  Permission is hereby granted to do whatever you like with
it, subject to the conditions that may exist in copyrighted materials. Should
you wish to make a contribution toward the improvement, modification, or
general performance of this module, please send us your comments:  why you
liked or disliked it, how you use it, and most important, how it helps your
work. We will receive your comments at avs@ncsc.org.

Please send AVS module bug reports to avs@ncsc.org.

******************************************************************************/
/*
 *  R E A D      P L O T 3 D  
 *
 *  AVS module reader for NASA plot3d files
 *
 *  Version 1.0   
 *   31 August    1994   Michael Rangitsch
 *
 *     Will only read multiblock files
 *
 *
 *          the preprocessor flags DEBUG and DEBUG1 print out more information 
 *          (lots of it) during the run
 *          
 */

#include <stdio.h>
#include <ctype.h>
#include <string.h>

#include <avs/util.h>
#include <avs/f_utils.h>
#include <avs/err.h>
#include <avs/om.h>
#include <avs/fld.h>
#include <avs/avs.h>
#include <avs/math.h>

#define  linelen  100
#define  true       1
#define  false      0

#define   DEBUG
#define   DEBUG0
#define   DEBUG1
#define   DEBUG2

#define READ_ERROR(str) ERRerror(str, 1, ERR_CHAIN, "read plot3D: ")

typedef  float   XYZ[3];

typedef  int     VERTS[8];

#define  CELL(nb,i,j,k) \
          ((blk_offset[(nb)]) + (jdim[(nb)]*idim[(nb)]*(k)) + \
           (idim[(nb)]*(j)) + (i))

#define  FACT8( bytes ) \
          (( ( ((bytes)/8) + 1) * 8 ))

char *labels[] = {"density","u-momentum","v-momentum", "w-momentum", "stagnation"};

char *units[] = {"kg/m**3","kgm**2/s**2","kgm**2/s**2","kgm**2/s**2","N/m**2"};


int read_plot3d_mb (char *XYZFileName, char *QFileName, int iblanks,
                    OMobj_id out);

/******************************
 *                            *
 * Read an PLOT3D data set    *
 *                            *
 ******************************/

/* Module description routine */

int DVread_PLOT3D_mb_update(OMobj_id elem_id)
{
	OMobj_id e, out;
	char *str1, *str2;
	int    iblanks, stat;

	e = OMfind_subobj(elem_id, OMstr_to_name("filename_xyz"), OM_OBJ_RD);
	if (OMget_str_val(e, &str1, 0) != 1) str1 = NULL;

	e = OMfind_subobj(elem_id, OMstr_to_name("filename_q"), OM_OBJ_RD);
	if (OMget_str_val(e, &str2, 0) != 1) str2 = NULL;

	e = OMfind_subobj(elem_id, OMstr_to_name("iblanks"), OM_OBJ_RD);
	OMget_int_val(e, &iblanks);

	out = OMfind_subobj(elem_id, OMstr_to_name("out"), OM_OBJ_RW);

	stat = read_plot3d_mb(str1, str2, iblanks, out);
	if (str1) free(str1);
	if (str2) free(str2);
	return(stat);
}

/* module compute section */

int read_plot3d_mb(char *XYZFileName, char *QFileName, int iblanks,
                   OMobj_id out)
{
	int             data_veclen, name_flag, util_flag, error_flag;
	int             cell_tsize, cell_veclen;
	int             node_csize, node_veclen;
	int             conn_list[8];
	int            *comp_list=NULL;

	FILE 	*xyzfile, *qfile;
	char  line[linelen],c,buffer[20],blockname[34],patchname[67];

	/* most of the variables in all caps are read in from the FLOW3D 
	   data set.  the deadint and dead flot variables are used to 
	   catch and discard information (like the cell face pointers and
	   cell face data values) which are not used to generate the ucd
	   structure.  the cell face variables may be useful later....   */

	int i, j, k, l, m, nvals, nlines;

	char  VARIABLENAME[28];

	int idata[10], nvar, nphase, npatch, nglue, ncells;
	int nun1, nnodes, ntime, nun2, nstored;

	int *ACTIVE_VARIABLES=NULL, *IBLK=NULL, *INFPCH=NULL, *INFGLU=NULL;
	int *LASTSET=NULL;
	int CYLINDRICAL=0;
	char first_word[linelen], dead_word[linelen], dead_word1[linelen];
	int deadint;

	float *XNN=NULL, *DATA=NULL;
	float deadfloat;

	float rad, theta, vrad, vtheta;
	int   node_numb, last_node_numb, iblk, *blk_offset=NULL;
	int   icell;
	VERTS *vertices=NULL;
	int   init_data;
	int   *node_index=NULL, *cell_used=NULL, *node_used=NULL;
	int   new_nodes, new_cells;

	float  min_extent[3], max_extent[3];
	float *node_minimum=NULL, *node_maximum=NULL;

	/* now define the parameters needed to remove the glued vertices */

	int   af,al,astep,bf,bl,bstep,cf,cl,cstep;
	int   i1,j1,k1,i2,j2,k2,iuse,juse,kuse;
	int   patch1,patch2,nwl1,nwl2,nblk1,nblk2;
	int   iptr,jptr,kptr;
	int   nrow1,ncol1;
	int   *cell1=NULL, *cell2=NULL, cell_no1, cell_no2, icelln;
	int   nrows,ncols,swd1,swd2;
	int   ni1,nj1,nk1,ni2,nj2,nk2;
	int   pointi[7],pointj[7],pointk[7];
	int   indices[7][4];
	int   iglue;
	int   vert1[4],vert2[4],tmp1,tmp2;
	int   lowi1,highi1,lowj1,highj1,lowk1,highk1;
	int   lowi2,highi2,lowj2,highj2,lowk2,highk2;
	int   switch_called[24],mcell;

	/* define the percent variables for status display */

	int   percent;
	char  message[30];

	/* define the geometry objects for internals */

	int            glues;

	char     name[10];
	int      *wall=NULL;
	int      *type=NULL;
	int      nintern, status;
	int      ifst,ilst,jfst,jlst,kfst,klst,niw;
	XYZ      verts[4];
	char     tname[40];

	/*  add additional parameters to read unformatted files */

	int      ftype;     /* flag for file type  0 unformatted 1 formatted */
	char     circ[2],tmp_char[1];
	int      fint,lint,lena,size;


	int      nblocks, nblocksq, inode, noffset, itmpnode;
	int      *idim=NULL,*jdim=NULL,*kdim=NULL, *tint=NULL;
	int      *idimq=NULL,*jdimq=NULL,*kdimq=NULL;
	int      *IBLANK=NULL;
	float    *tfloat=NULL, *Q=NULL;
	float    fsmach, alpha, re, time;

	/* EXPRESS */
	OMobj_id          elm, *fields;
	int dims[3], nspace, ndim, ncomp, veclen, dtype, stat;
	float *coord, *node_data;
	float min, max, rmin[10], rmax[10];


	/* only execute if a valid file names are passed */

	if ( XYZFileName == NULL || QFileName == NULL )
		return(1);

	if(( (xyzfile = FILEfopen(XYZFileName,"r")) == (FILE *) NULL ) ||
	   ( (qfile = FILEfopen(QFileName,"r")) == (FILE *) NULL)          ) {
		printf("Couldn't open %s or %s for reading.\n", XYZFileName, QFileName);
		return(0);
	}
	/* input the file header */
#ifdef DEBUG_PRINT
	fprintf(stderr, "PLOT3D read starting\n");
#endif

	/* check to see if we have a formatted or unformatted file */

	ftype = 0;

	fread( &fint,sizeof(int),1,xyzfile);

	if( fint > 10 ) ftype = 1;

#ifdef DEBUG_PRINT
	printf("ftype = %d\n",ftype);
#endif

	rewind(xyzfile);

	if( ftype )
	{
		fscanf(xyzfile,"%d",&nblocks);
	}
	else
	{
		fread( &fint,sizeof(int),1,xyzfile);
		fread( &nblocks,sizeof(int),1,xyzfile);
		fread( &lint,sizeof(int),1,xyzfile);
	}

#ifdef DEBUG_PRINT
	printf("nblocks = %d\n",nblocks);
#endif

	elm = OMfind_subobj(out, OMstr_to_name("nblocks"), OM_OBJ_RD);
	OMset_int_val(elm, nblocks);

	if (nblocks == 0)
		return(0);

	fields = (OMobj_id *)malloc(nblocks*sizeof(OMobj_id));

	elm = OMfind_subobj(out, OMstr_to_name("fields"), OM_OBJ_RD);
	for (i=0; i< nblocks; i++)
		if ((stat = OMget_array_val(elm, i, fields+i, OM_OBJ_RD)) != 1)
			return(stat);


	/*  now allocate and read in the dimensions array */


	idim = (int *)malloc( FACT8(nblocks*sizeof(int)) );

	jdim = (int *)malloc( FACT8(nblocks*sizeof(int)) );

	kdim = (int *)malloc( FACT8(nblocks*sizeof(int)) );

	if( ftype ) {
		for( i = 0; i < nblocks; i++ )	{
			fscanf(xyzfile,"%d %d %d",&idim[i],&jdim[i],&kdim[i]);
#ifdef DEBUG_PRINT
			printf(" block %d, ni %d, nj %d, nk %d\n",i,idim[i],jdim[i],kdim[i]);
#endif
		}
	}
	else {

		/* IAC CODE CHANGE :      if( tint ) free(tint); */
		if( tint )  free(tint);
		tint = (int *)malloc( FACT8((3*nblocks+2)*sizeof(int)) );
		fread( tint,sizeof(int),3*nblocks+2,xyzfile);
		j = 1;
		for( i = 0; i < nblocks; i++ ) {
			idim[i] = tint[j];
			jdim[i] = tint[j+1];
			kdim[i] = tint[j+2];
#ifdef DEBUG_PRINT
			printf(" block %d, ni %d, nj %d, nk %d\n ",i,idim[i],jdim[i],kdim[i]);
#endif
			j = j + 3;
		}
	}

	/* now set the sizes and read in the coordinate arrays */

	nnodes = 0;
	nspace = 3;
	ndim = 3;
	for( i = 0; i < nblocks; i++) {
		nnodes = nnodes + idim[i]*jdim[i]*kdim[i];
		if (FLDset_nspace(fields[i], nspace) != 1) {
			READ_ERROR("Error setting nspace");
		}
		if (FLDset_ndim(fields[i], ndim) != 1) {
			READ_ERROR("Error setting ndim");
		}
		dims[0] = idim[i];
		dims[1] = jdim[i];
		dims[2] = kdim[i];
		if (FLDset_dims (fields[i], dims) != 1) {
			READ_ERROR("Error setting dims");
		}
	}
	if(iblanks)
		IBLANK = (int *)malloc( FACT8(nnodes*sizeof(int)) );

	if( ftype ) {
		noffset = 0;
		for( l = 0; l < nblocks; l++ )	{
			if (FLDget_coord(fields[l],(float **)&coord, &size, OM_GET_ARRAY_WR) != 1) {
				READ_ERROR("Error setting coordinate array");
			}
			for(i = 0; i < idim[l]*jdim[l]*kdim[l]; i++ )
				fscanf( xyzfile, "%f", coord+i*nspace );

			for(i = 0; i < idim[l]*jdim[l]*kdim[l]; i++ )
				fscanf( xyzfile, "%f",  coord+i*nspace+1 );

			for(i = 0; i < idim[l]*jdim[l]*kdim[l]; i++ )
				fscanf( xyzfile, "%f", coord+i*nspace+2 );

			if( iblanks ) {
				for(i = 0; i < idim[l]*jdim[l]*kdim[l]; i++ )
					fscanf( xyzfile, "%d", &IBLANK[noffset+i] );
			}
			noffset = noffset + idim[l]*jdim[l]*kdim[l];
			ARRfree(coord);
		}
	}
	else  {
		noffset = 0;
		inode = 0;

		for( l = 0; l < nblocks; l++ )	{
			if (FLDget_coord(fields[l],(float **)&coord, &size, OM_GET_ARRAY_WR) != 1) {
				READ_ERROR("Error setting coordinate array");
			}

			/* IAC CODE CHANGE :       if(tfloat) free(tfloat); */
			if(tfloat)  free(tfloat);
			tfloat = (float *)malloc( FACT8((4*idim[l]*jdim[l]*kdim[l]+1)
							*sizeof(float)) );

			if(iblanks) {
				fread(tfloat,sizeof(float),4*idim[l]*jdim[l]*kdim[l]+2,xyzfile);
				for ( i = 0; i < idim[l]*jdim[l]*kdim[l]; i++ ) {
					coord[i*nspace]   = tfloat[i+1];
					coord[i*nspace+1] = tfloat[i+1+idim[l]*jdim[l]*kdim[l]];
					coord[i*nspace+2] = tfloat[i+1+2*idim[l]*jdim[l]*kdim[l]];
					IBLANK[i+noffset] = (int)tfloat[i+1+3*idim[l]*jdim[l]*kdim[l]];
				}
				noffset = noffset + idim[l]*jdim[l]*kdim[l];
			}
			else {
				fread(tfloat,sizeof(float),3*idim[l]*jdim[l]*kdim[l]+2,xyzfile);
				for ( i = 0; i < idim[l]*jdim[l]*kdim[l]; i++ ) {
					coord[i*nspace]   = tfloat[i+1];
					coord[i*nspace+1] = tfloat[i+1+idim[l]*jdim[l]*kdim[l]];
					coord[i*nspace+2] = tfloat[i+1+2*idim[l]*jdim[l]*kdim[l]];
					IBLANK[i+noffset] = (int)tfloat[i+1+3*idim[l]*jdim[l]*kdim[l]];
				}
				noffset = noffset + idim[l]*jdim[l]*kdim[l];
			}
			ARRfree(coord);
			if(tfloat)  {
				free(tfloat);
				tfloat = NULL;
			}
		}
	}

	/* now read in the Q file */

	ftype = 0;

	fread( &fint,sizeof(int),1,qfile);

	if( fint > 10 ) ftype = 1;

#ifdef DEBUG_PRINT
	printf("ftype = %d\n",ftype);
#endif

	rewind(qfile);

	if( ftype ) {
		fscanf(qfile,"%d",&nblocksq);
	}
	else {
		fread( &fint,sizeof(int),1,qfile);
		fread( &nblocksq,sizeof(int),1,qfile);
		fread( &lint,sizeof(int),1,qfile);
	}

#ifdef DEBUG_PRINT
	printf("nblocksq = %d\n",nblocksq);
#endif

	/*  now allocate and read in the dimensions array */


	idimq = (int *)malloc( FACT8(nblocksq*sizeof(int)) );

	jdimq = (int *)malloc( FACT8(nblocksq*sizeof(int)) );

	kdimq = (int *)malloc( FACT8(nblocksq*sizeof(int)) );

	if( ftype )
	{
		for( i = 0; i < nblocksq; i++ )
		{
			fscanf(qfile,"%d %d %d",&idimq[i],&jdimq[i],&kdimq[i]);
#ifdef DEBUG_PRINT
			printf(" block %d, ni %d, nj %d, nk %d\n",i,idimq[i],jdimq[i],kdimq[i]);
#endif
		}
	}
	else {
		/* IAC CODE CHANGE :      if( tint ) free(tint); */
		if( tint )  free(tint);
		tint = (int *)malloc( FACT8((3*nblocksq+2)*sizeof(int)) );
		fread( tint,sizeof(int),3*nblocksq+2,qfile);
		j = 1;
		for( i = 0; i < nblocksq; i++ )
		{
			idimq[i] = tint[j];
			jdimq[i] = tint[j+1];
			kdimq[i] = tint[j+2];
#ifdef DEBUG_PRINT
			printf(" block %d, ni %d, nj %d, nk %d\n ",i,idimq[i],jdimq[i],kdimq[i]);
#endif
			j = j + 3;
		}
	}

	/* check for file consistency */

	if( nblocks != nblocksq )  {
		printf(" xyz and q files have different block counts\n");
		return(0);
	}

	for( l = 0; l < nblocks; l++ )
	{
		if( idim[l] != idimq[l] ||
		   jdim[l] != jdimq[l] ||
		   kdim[l] != kdimq[l] )
		{
			printf(" xyz and q files are inconsistent\n");
			return(0);
		}
	}

	/** DATA **/
	ncomp = 5;
	veclen = 1;
	dtype = DTYPE_FLOAT;
	if( ftype ) {
		for ( l = 0; l < nblocks; l++ )	{
			if (FLDset_node_data_ncomp (fields[l], ncomp) != 1) {
				READ_ERROR("Error setting nnode_data");
			}
			fscanf(qfile,"%f %f %f %f",&fsmach,&alpha,&re,&time);
			for( m = 0; m < ncomp; m++ ) {
				if ( FLDset_node_data_comp (fields[l], m, veclen, labels[m], units[m]) != 1) {
					READ_ERROR("error setting node data component");
				}
				if (FLDget_node_data(fields[l], m, &dtype, (char **)&node_data, &size,
						     OM_GET_ARRAY_WR) != 1) {
					READ_ERROR("Error getting node data");
				}
				inode = 0;
				for( k = 0; k < kdimq[l]; k++ )
					for( j = 0; j < jdimq[l]; j++ )
						for( i = 0; i < idimq[l]; i++ )
							fscanf(qfile,"%f",&node_data[inode++]);
				ARRfree(node_data);
			}
		}
	}
	else {
		inode = 0;
		noffset = 0;
		for( l = 0; l < nblocks; l++ ) 	{
			if (FLDset_node_data_ncomp (fields[l], ncomp) != 1) {
				READ_ERROR("Error setting nnode_data");
			}

			/* IAC CODE CHANGE :       if(tfloat) free(tfloat); */
			if(tfloat)  free(tfloat);
			tfloat = (float *)malloc( FACT8((4+2)*sizeof(float)) );
			fread(tfloat,sizeof(float),6,qfile);
#ifdef DEBUG_PRINT
			printf("%f %f %f %f %f %f\n",tfloat[0],tfloat[1],tfloat[2],tfloat[3],tfloat[4],tfloat[5]);
#endif

			/* IAC CODE CHANGE :       if(tfloat) free(tfloat); */
			if(tfloat)  free(tfloat);
			tfloat = (float *)malloc( FACT8((5*idimq[l]*jdimq[l]*kdimq[l]+2)
							*sizeof(float)) );

			fread(tfloat,sizeof(float),5*idimq[l]*jdimq[l]*kdimq[l]+2,qfile);

			itmpnode = 0;
			for( m = 0; m < ncomp; m++ ) {
				if ( FLDset_node_data_comp (fields[l], m, veclen, labels[m], units[m]) != 1) {
					READ_ERROR("error setting node data component");
				}
				if (FLDget_node_data(fields[l], m, &dtype, (char **)&node_data, &size,
						     OM_GET_ARRAY_WR) != 1) {
					READ_ERROR("Error getting node data");
				}
				for( i = 0; i < idimq[l]*jdimq[l]*kdimq[l]; i++ )
					node_data[i] = tfloat[i+1+m*idimq[l]*jdimq[l]*kdimq[l]];
				ARRfree(node_data);
			}
			noffset = noffset + idim[l]*jdim[l]*kdim[l];
			if(tfloat) {
				free(tfloat);
				tfloat = NULL;
			}
		}
	}
	for ( l = 0; l < nblocks; l++ )	{
		if (FLDget_node_data_ncomp (fields[l], &ncomp) != 1) {
			READ_ERROR("Error setting nnode_data");
		}
		for( m = 0; m < ncomp; m++ ) {
			if (FLDget_node_data_minmax(fields[l], m, (char *)&min, (char *)&max) != 1) {
				READ_ERROR("Error getting node minmax data");
			}
			if (l==0) {
				rmin[m] = min;
				rmax[m] = max;
			}
			else {
				if (min < rmin[m])
					rmin[m] = min;
				if (max > rmax[m])
					rmax[m] = max;
			}
		}
	}
	for ( l = 0; l < nblocks; l++ )	{
		if (FLDget_node_data_ncomp (fields[l], &ncomp) != 1) {
			READ_ERROR("Error setting nnode_data");
		}
		for( m = 0; m < ncomp; m++ ) {
			if (FLDset_node_data_minmax(fields[l], m,
						    (char *)&rmin[m], (char *)&rmax[m], DTYPE_FLOAT) != 1) {
				READ_ERROR("Error setting minmax data");
			}
		}
	}
	fclose (xyzfile);
	fclose (qfile);

	/* end of data read */

#ifdef DEBUG_PRINT
	printf(" got past the data input \n");
#endif

	free((char *)fields);
	free((char *)idim);
	free((char *)idimq);
	free((char *)jdim);
	free((char *)jdimq);
	free((char *)kdim);
	free((char *)kdimq);
	if (IBLANK) free((char *)IBLANK);
	if (tint) free((char *)tint);

	return(1);
}
